use crate::qso::Qso;
use crate::station::Station;

use adif::{AdifFile, AdifType};
use indexmap::indexmap;
use std::fs;

pub trait AdifEncoder {
	fn encode(station: &Station, qsos: Vec<Qso>) -> AdifFile;
	fn save(station: &Station, qsos: Vec<Qso>, filename: &str) -> Result<(), String> {
		let adif = Self::encode(station, qsos);
		let contents = adif.serialize().map_err(|x| format!("{x}"))?;
		fs::write(filename, contents).map_err(|x| format!("{x}"))?;

		Ok(())
	}
}

pub struct QrzEncoder;

impl AdifEncoder for QrzEncoder {
	fn encode(station: &Station, qsos: Vec<Qso>) -> AdifFile {
		let header = indexmap! {
			"station_callsign" => AdifType::Str(station.callsign.clone()),
			"adif_ver" => AdifType::Str("3.0.4".to_string()),
			"programid" => AdifType::Str("limalogger".to_string()),
			"programversion" => AdifType::Str("v0.3.0".to_string()),
			"created_timestamp" => AdifType::Str(
				format!("{}", chrono::Utc::now().format("%Y%m%d %H%M%S"))
			)
		}
		.into();

		let body = qsos
			.into_iter()
			.map(|qso| {
				let band = qso.calc_band();

				let mut hsh = indexmap! {
					"qso_date" => AdifType::Date(qso.time.date_naive()),
					"time_off" => AdifType::Time(qso.time.time()),
					"time_on" => AdifType::Time(qso.time.time()),
					"freq" => AdifType::Number(qso.freq.parse::<f64>().unwrap_or(0.0) / 1000.0),
					"mode" => AdifType::Str(qso.mode),
					"tx_pwr" => AdifType::Number(qso.power),
					"call" => AdifType::Str(qso.callsign),
					"qth" => AdifType::Str(qso.qth),
					"rst_sent" => AdifType::Str(qso.rst_sent),
					"rst_rcvd" => AdifType::Str(qso.rst_recv),
					"operator" => AdifType::Str(station.callsign.clone()),
				};

				if let Some(band) = band {
					hsh.insert("band", AdifType::Str(band.to_string()));
				}

				hsh.into()
			})
			.collect();

		AdifFile { header, body }
	}
}
