use crate::adif::{AdifEncoder, QrzEncoder};
use crate::qso::Qso;
use crate::station::Station;
use crate::traits::TuiValue;

use chrono::Utc;
use serde::{Deserialize, Serialize};
use std::fs;
use std::io::{stdin, Write};
use term_table::{TableBuilder, TableStyle};

const MAX_COLUMN_WIDTH: usize = 40;

#[derive(Serialize, Deserialize, Default)]
pub struct State {
	qsos: Vec<Qso>,
	#[serde(default)]
	station: Station,
}

impl State {
	fn sort(&mut self) {
		self.qsos.sort_by(|a, b| a.time.cmp(&b.time));
	}

	fn load_file(filename: &str) -> anyhow::Result<Self> {
		let s = fs::read_to_string(filename)?;

		Ok(toml::from_str(&s).expect("Could not parse file"))
	}

	fn save_file(&self, filename: &str) -> anyhow::Result<()> {
		let s = toml::to_string(&self)?;
		fs::write(filename, s)?;

		Ok(())
	}
}

pub struct App {
	state: State,
	filename: String,
}

impl App {
	pub fn new(filename: String) -> Option<Self> {
		let state = match State::load_file(&filename) {
			Ok(x) => x,
			Err(_) => {
				// Create new file
				let state = State::default();
				if let Err(e) = state.save_file(&filename) {
					eprintln!("Could not create file {filename}: {e}");
					return None;
				}
				state
			}
		};

		Some(App { state, filename })
	}

	pub fn run(&mut self) {
		loop {
			self.render_table();
			println!(
				"\t[A] Add QSO\t[E] Edit QSO\t[D] Del QSO\t[C] Edit Station Callsign\t[X] Export ADIF File\t[Q] Quit"
			);

			// Wish this had unicode support ):
			let c = getch::Getch::new().getch().unwrap() as char;
			match c {
				'a' | 'A' => {
					self.add_qso();
				}
				'e' | 'E' => {
					self.edit_qso();
				}
				'd' | 'D' => {
					self.del_qso();
				}
				'c' | 'C' => {
					self.edit_station_callsign();
				}
				'x' | 'X' => {
					self.export_adif_file();
				}
				'q' | 'Q' => break,
				_ => {}
			}

			self.state.sort();

			if let Err(e) = self.state.save_file(&self.filename) {
				eprintln!("Couldn't save file: {e}");
			}
		}
	}

	fn render_table(&self) {
		let mut rows = vec![Qso::table_header()];
		rows.extend(
			self.state
				.qsos
				.iter()
				.enumerate()
				.map(|(i, qso)| qso.to_row(i)),
		);

		let mut table = TableBuilder::new()
			.style(TableStyle::extended())
			.rows(rows)
			.build();
		table.max_column_width = MAX_COLUMN_WIDTH;

		clearscreen::clear().unwrap();

		// Header
		println!("Station: {}", self.state.station.callsign);
		// Table
		print!("{}", table.render());
	}

	fn add_qso(&mut self) {
		let (freq, mode, power) = self
			.state
			.qsos
			.last()
			.map_or((String::new(), String::new(), 100.0), |qso| {
				(qso.freq.clone(), qso.mode.clone(), qso.power)
			});

		let qso = Qso {
			time: Utc::now(),
			freq,
			mode,
			power,
			callsign: String::new(),
			qth: String::new(),
			rst_sent: String::new(),
			rst_recv: String::new(),
			comments: String::new(),
		};

		if let Some(qso) = self.modify_qso(qso, self.state.qsos.len()) {
			self.state.qsos.push(qso);
		}
	}

	fn edit_qso(&mut self) {
		print!("Id [Blank to cancel]? ");
		Self::flush();
		let mut buffer = String::new();
		stdin().read_line(&mut buffer).unwrap();
		let buffer = buffer.trim();
		if let Ok(id) = buffer.parse::<usize>() {
			if id < self.state.qsos.len() {
				let qso = self.state.qsos[id].clone();
				if let Some(qso) = self.modify_qso(qso, id) {
					self.state.qsos[id] = qso;
				}
			}
		}
	}

	fn del_qso(&mut self) {
		print!("Id [Blank to cancel]? ");
		Self::flush();
		let mut buffer = String::new();
		stdin().read_line(&mut buffer).unwrap();
		let buffer = buffer.trim();
		if let Ok(id) = buffer.parse() {
			if id < self.state.qsos.len() {
				self.state.qsos.remove(id);
			}
		}
	}

	fn edit_station_callsign(&mut self) {
		self.state.station.callsign =
			Self::prompt("Station Callsign", self.state.station.callsign.clone()).to_uppercase();
	}

	fn export_adif_file(&self) {
		print!("File name [Blank to cancel]? ");
		Self::flush();
		let mut buffer = String::new();
		stdin().read_line(&mut buffer).unwrap();
		let buffer = buffer.trim();
		if !buffer.is_empty() {
			if let Err(e) = QrzEncoder::save(&self.state.station, self.state.qsos.clone(), buffer) {
				println!("Could not export ADIF file: {e}\nPress any key to continue...",);
				getch::Getch::new().getch().unwrap();
			}
		}
	}

	fn modify_qso(&mut self, mut qso: Qso, id: usize) -> Option<Qso> {
		loop {
			clearscreen::clear().unwrap();
			self.render_table();

			let mut table = TableBuilder::new()
				.style(TableStyle::extended())
				.rows(vec![
					Qso::table_header_with_suffixes(&[
						"", "[T]", "[F]", "[M]", "[P]", "[C]", "[Q]", "[E]", "[R]", "[O]",
					]),
					qso.to_row(id),
				])
				.build();
			table.max_column_width = MAX_COLUMN_WIDTH;
			println!("{}\t[S] Save\t[X] Cancel", table.render());

			let c = getch::Getch::new().getch().unwrap() as char;
			match c {
				't' | 'T' => {
					qso.time = Self::prompt("Time & Date", qso.time);
				}
				'f' | 'F' => {
					qso.freq = Self::prompt("Frequency (kHz)", qso.freq);
				}
				'm' | 'M' => {
					qso.mode = Self::prompt("Mode", qso.mode).to_uppercase();
				}
				'p' | 'P' => {
					qso.power = Self::prompt("Power", qso.power);
				}
				'c' | 'C' => {
					qso.callsign = Self::prompt("Callsign", qso.callsign).to_uppercase();
				}
				'q' | 'Q' => {
					qso.qth = Self::prompt("QTH", qso.qth);
				}
				'e' | 'E' => {
					qso.rst_sent = Self::prompt("RST Sent", qso.rst_sent);
				}
				'r' | 'R' => {
					qso.rst_recv = Self::prompt("RST Recv", qso.rst_recv);
				}
				'o' | 'O' => {
					qso.comments = Self::prompt("Comments", qso.comments);
				}
				's' | 'S' => {
					break Some(qso);
				}
				'x' | 'X' => {
					break None;
				}
				_ => {}
			}
		}
	}

	fn prompt<T: TuiValue>(label: &str, default: T) -> T {
		println!();

		let default_s: String = default.to_string();
		let default_s = default_s.trim();

		if default_s.trim().is_empty() {
			print!("{label}? ");
		} else {
			print!("{label} [{default_s}]? ");
		}
		Self::flush();

		let mut buffer = String::new();
		stdin().read_line(&mut buffer).unwrap();
		let buffer = buffer.trim();
		if buffer.is_empty() {
			default
		} else {
			match T::from_str(buffer) {
				Some(x) => x,
				// Invalid value - prompt again
				None => Self::prompt(label, default),
			}
		}
	}

	fn flush() {
		std::io::stdout().flush().unwrap();
	}
}
