use chrono::{DateTime, Utc};
use serde::{Deserialize, Serialize};
use term_table::row::Row;
use term_table::table_cell::TableCell;

#[derive(Clone, Serialize, Deserialize)]
pub struct Qso {
	pub time: DateTime<Utc>,
	pub freq: String,
	pub mode: String,
	pub power: f64,
	pub callsign: String,
	pub qth: String,
	pub rst_sent: String,
	pub rst_recv: String,
	pub comments: String,
}

impl Qso {
	pub fn table_header() -> Row<'static> {
		Self::table_header_with_suffixes(&["", "", "", "", "", "", "", "", "", ""])
	}

	pub fn table_header_with_suffixes(suffixes: &[&str; 10]) -> Row<'static> {
		const COL_NAMES: [&str; 10] = [
			"id",
			"Time & Date",
			"Freq (kHz)",
			"Mode",
			"Power (W)",
			"Callsign",
			"QTH",
			"RST Sent",
			"RST Recv",
			"Comments",
		];
		Row::new(
			COL_NAMES
				.iter()
				.enumerate()
				.map(|(i, col_name)| TableCell::new(format!("{} {}", col_name, suffixes[i]))),
		)
	}

	pub fn time_s(&self) -> String {
		self.time.format("%a %b %e %Y %T").to_string()
	}

	/// gets the frequency band from the current frequency.
	pub fn calc_band(&self) -> Option<&'static str> {
		// freq is in kHz
		let freq: f64 = self.freq.parse().ok()?;

		// see https://www.adif.org/304/ADIF_304.htm#Band_Enumeration
		let band = match freq {
			x if (136.0..137.0).contains(&x) => "2190m",
			x if (472.0..479.0).contains(&x) => "630m",
			x if (501.0..504.0).contains(&x) => "560m",
			x if (1_800.0..2_000.0).contains(&x) => "160m",
			x if (3_500.0..4_000.0).contains(&x) => "80m",
			x if (5_102.0..5_406.5).contains(&x) => "60m",
			x if (7_000.0..7_300.0).contains(&x) => "40m",
			x if (10_000.0..10_150.0).contains(&x) => "30m",
			x if (14_000.0..14_350.0).contains(&x) => "20m",
			x if (18_068.0..18_168.0).contains(&x) => "17m",
			x if (21_000.0..21_450.0).contains(&x) => "15m",
			x if (24_890.0..24_990.0).contains(&x) => "12m",
			x if (28_000.0..29_700.0).contains(&x) => "10m",
			x if (50_000.0..54_000.0).contains(&x) => "6m",
			x if (70_000.0..71_000.0).contains(&x) => "4m",
			x if (144_000.0..148_000.0).contains(&x) => "2m",
			x if (222_000.0..225_000.0).contains(&x) => "1.25m",
			x if (420_000.0..450_000.0).contains(&x) => "70cm",
			x if (902_000.0..928_000.0).contains(&x) => "33cm",
			x if (1_240_000.0..1_300_000.0).contains(&x) => "23cm",
			x if (2_300_000.0..2_450_000.0).contains(&x) => "13cm",
			x if (3_300_000.0..3_500_000.0).contains(&x) => "9cm",
			x if (5_650_000.0..5_925_000.0).contains(&x) => "6cm",
			x if (10_000_000.0..10_500_000.0).contains(&x) => "3cm",
			x if (24_000_000.0..24_250_000.0).contains(&x) => "1.25cm",
			x if (47_000_000.0..47_200_000.0).contains(&x) => "6mm",
			x if (75_500_000.0..81_000_000.0).contains(&x) => "4mm",
			x if (119_980_000.0..120_020_000.0).contains(&x) => "2.5mm",
			x if (142_000_000.0..149_000_000.0).contains(&x) => "2mm",
			x if (241_000_000.0..250_000_000.0).contains(&x) => "1mm",
			_ => return None,
		};

		Some(band)
	}

	pub fn to_row(&self, index: usize) -> Row<'_> {
		Row::new([
			TableCell::new(index),
			TableCell::new(self.time_s()),
			TableCell::new(&self.freq),
			TableCell::new(&self.mode),
			TableCell::new(self.power),
			TableCell::new(&self.callsign),
			TableCell::new(&self.qth),
			TableCell::new(&self.rst_sent),
			TableCell::new(&self.rst_recv),
			TableCell::new(&self.comments),
		])
	}
}
